from ...const import __api_url__

__all__ = ("Route",)

V1_URL = "https://www.anilibria.tv/public"
V3_URL = f"https://{__api_url__}"


class Route:
    def __init__(self, method: str, endpoint: str, *, is_v1: bool = False):
        self.method: str = method
        self.endpoint: str = endpoint
        self._is_v1: bool = is_v1

    @property
    def url(self) -> str:
        url = V1_URL if self._is_v1 else V3_URL
        return f"{url}{self.endpoint}"



from typing import List, Type

from cattrs import Converter

from .title import Episode, RutubeEpisode

__all__ = ("converter",)

converter = Converter()


def _playlist_hook(data: dict | list, type_: Type[Episode | RutubeEpisode]):
    if isinstance(data, list):
        return [converter.structure(_, type_) for _ in data]
    if isinstance(data, dict):
        return {k: converter.structure(v, type_) for k, v in data.items()}


def _series_hook(data: dict | list, type_: type):
    return _playlist_hook(data, Episode)


def _rutube_series_hook(data: dict | list, type_: type):
    return _playlist_hook(data, RutubeEpisode)


converter.register_structure_hook(dict[str, Episode] | List[Episode], _series_hook)
converter.register_structure_hook(
    dict[str, RutubeEpisode] | List[RutubeEpisode], _rutube_series_hook
)



from enum import Enum
from typing import Optional

from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override

from ...models import Episode, Player, Title, Torrents
from ...models.attrs_utils import define
from ...models.cattrs_utils import converter
from .base import BaseEvent
from .internal import Connect, TitleEpisode

__all__ = (
    "EncodeStart",
    "EncodeProgress",
    "EncodeEnd",
    "EncodeFinish",
    "PlaylistUpdate",
    "TitleUpdate",
    "TorrentUpdate",
    "Subscription",
    "EventType",
)


@define()
class _BaseEncodeEvent(BaseEvent):
    """
    Базовый объект для события кодирования
    """

    id: str
    "ID тайтла"
    episode: str
    "Номер эпизода"


@define()
class _EncodeEvent(_BaseEncodeEvent):
    resolution: str
    "Разрешение загружаемого эпизода"
    quality: str
    "Качество загружаемого эпизода"


@define()
class EncodeStart(_EncodeEvent):
    """
    Объект для события ``on_encode_start``.
    Вызывается, когда начинается загрузка эпизода на сервер.

    .. code-block:: python

       @client.on(EncodeStart)
       async def start_encode(event: EncodeStart):
           ...
    """

    is_reupload: bool
    "Является ли загрузка эпизода перезаливом"


@define()
class EncodeProgress(_EncodeEvent):
    """
    Объект для события ``on_encode_progress``.
    Вызывается на каждые 5% загрузки.

    .. code-block:: python

       @client.on(EncodeProgress)
       async def progress(event: EncodeProgress):
           ...
    """

    encoded_percent: str
    "Текущий процент загрузки"


@define()
class EncodeEnd(_EncodeEvent):
    """
    Объект для события ``on_encode_end``.
    Вызывается, когда было загружено одно качество эпизода на сервер.

    .. code-block:: python

       @client.on(EncodeEnd)
       async def encode_end(event: EncodeEnd):
           ...
    """

    ...


@define()
class EncodeFinish(_BaseEncodeEvent):
    """
    Объект для события ``on_encode_finish``.
    Вызывается, когда все возможные качества эпизода были загружены на сервер.

    .. code-block:: python

       @client.on(EncodeFinish)
       async def encode_finish(event: EncodeFinish):
           ...
    """

    ...


@define()
class PlaylistUpdate(BaseEvent):
    """
    Модель для ивента ``on_playlist_update``.
    Вызывается при обновлении данных плейлиста тайтла.

    .. code-block:: python

      @client.on(PlaylistUpdate)
      async def on_playlist_update(event: PlaylistUpdate):
          ...
    """

    id: Optional[int] = None
    "ID тайтла"
    player: Player | None = None
    "Обновлённый плейлист"
    updated_episode: Episode | None = None
    "Обновлённый/добавленный эпизод"
    episode: str | None = None
    "Номер эпизода"
    diff: dict | None = None
    "Словарь с предыдущими значениями"
    reupload: bool | None = None
    "Является ли это перезаливом"


@define()
class TitleUpdate(BaseEvent):
    """
    Модель для ивента ``on_title_update``.
    Вызывается при изменении информации о тайтле.

    .. code-block:: python

      @client.on(TitleUpdate)
      async def on_title_update(event: TitleUpdate):
          ...
    """

    title: Title = None
    "Объект тайтла"
    diff: dict = None
    "Предыдущие значения тайтла"


@define()
class TorrentUpdate(BaseEvent):
    """
    Модель для ивента `on_torrent_update`.
    Вызывается при изменении информации о торрент файлах.

    .. code-block:: python

      @client.on(TorrentUpdate)
      async def on_torrent_update(event: TorrentUpdate):
          ...
    """

    id: str = None
    "ID тайтла"
    torrents: Torrents = None
    "Информация о торренте"
    updated_torrent_id: int = None
    "ID обновлённого торрента"
    diff: dict = None
    "Предыдущие значения"


@define()
class Subscription(BaseEvent):
    """
    Модель для ивента `on_subscription`.
    Вызывается при подписке на события, приходящие с websocket.

    .. code-block:: python

      @client.on(Subscription)
      async def event_subscription(event: Subscription):
          ...
    """
    subscribe: str
    "Статус подписки"
    subscription_id: int
    "ID подписки"


class EventType(Enum):
    """
    Обозначает ивенты, которые принимает Websocket
    """

    # Anilibria events
    TITLE_UPDATE = TitleUpdate
    PLAYLIST_UPDATE = PlaylistUpdate
    ENCODE_START = EncodeStart
    ENCODE_PROGRESS = EncodeProgress
    ENCODE_END = EncodeEnd
    ENCODE_FINISH = EncodeFinish
    TORRENT_UPDATE = TorrentUpdate
    SUBSCRIPTION = Subscription

    # Internal events
    TITLE_EPISODE = TitleEpisode
    CONNECT = Connect


# Hooks
# TODO: Find a better place
unstruct_hook = make_dict_unstructure_fn(
    EncodeStart, converter, is_reupload=override(rename="isReupload")
)
struct_hook = make_dict_structure_fn(
    EncodeStart, converter, is_reupload=override(rename="isReupload")
)

converter.register_unstructure_hook(EncodeStart, unstruct_hook)
converter.register_structure_hook(EncodeStart, struct_hook)



from ...utils import dict_filter_none
from .request import Request
from .route import Route

__all__ = ("PublicRequest",)


class PublicRequest(Request):
    def __init__(
        self,
        proxy: str | None = None,
    ) -> None:
        super().__init__(proxy)

    # v1
    """
    Представляет собой запросы на сайт "https://www.anilibria.tv/public"
    Здесь не будут реализованы методы для "/api", так как существует v2, в котором точно такие же методы,
    и к тому же документация к "/api" не обновлялась более двух лет.
    """

    async def login(self, mail: str, password: str) -> dict:
        payload: dict = {"mail": mail, "passwd": password}
        route = Route("POST", "/login.php", is_v1=True)

        return await self.request(route, data=payload)

    # v3

    async def get_title(
        self,
        id: int | None = None,
        code: str | None = None,
        torrent_id: int | None = None,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            id=id,
            code=code,
            torrent_id=torrent_id,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        return await self.request(Route("GET", "/title"), payload)

    async def get_titles(
        self,
        id_list: list[int] | None = None,
        code_list: list[str] | None = None,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            id_list=id_list,
            code_list=code_list,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/title/list"), payload)

    async def get_updates(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        since: int | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        after: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/title/updates"), payload)

    async def get_changes(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        since: int | None = None,
        description_type: str | None = None,
        after: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/title/changes"), payload)

    async def get_schedule(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        days: list[str] = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
    ) -> list[dict]:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            days=days,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        return await self.request(Route("GET", "/title/schedule"), payload)

    async def get_random_title(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        return await self.request(Route("GET", "/title/random"), payload)

    async def get_youtube(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        since: int | None = None,
        after: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/youtube"), payload)

    async def get_feed(
        self,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        since: int | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        after: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/feed"), payload)

    async def get_years(self) -> list[int]:
        return await self.request(Route("GET", "/years"))

    async def get_genres(self, sorting_type: int = 0) -> list[str]:
        payload: dict = dict_filter_none(
            sorting_type=sorting_type,
        )
        return await self.request(Route("GET", "/genres"), payload)

    async def get_team(self) -> dict:
        return await self.request(Route("GET", "/team"))

    async def get_seed_stats(
        self,
        users: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        after: int | None = None,
        sort_by: str | None = None,
        order: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            users=users,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            sort_by=sort_by,
            order=order,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/torrent/seed_stats"), payload)

    async def get_rss(
        self,
        rss_type: str,
        session: str,
        since: int | None = None,
        after: int | None = None,
        limit: int | None = None,
    ) -> str:
        payload: dict = dict_filter_none(
            rss_type=rss_type, session=session, since=since, after=after, limit=limit
        )
        return await self.request(Route("GET", "/torrent/rss"), payload)

    async def search_titles(
        self,
        search: list[str] | None = None,
        year: list[str] | None = None,
        season_code: list[str] | None = None,
        genres: list[str] | None = None,
        team: list[str] | None = None,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        after: int | None = None,
        limit: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            search=search,
            year=year,
            season_code=season_code,
            genres=genres,
            team=team,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/title/search"), payload)

    async def advanced_search(
        self,
        query: str,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
        after: int | None = None,
        order_by: str | None = None,
        limit: int | None = None,
        sort_direction: int | None = None,
        page: int | None = None,
        items_per_page: int | None = None,
    ) -> dict:
        payload: dict = dict_filter_none(
            query=query,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            order_by=order_by,
            limit=limit,
            sort_direction=sort_direction,
            page=page,
            items_per_page=items_per_page,
        )
        return await self.request(Route("GET", "/title/search/advanced"), payload)

    async def get_user(
        self,
        session: str,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
    ) -> dict:
        payload = dict_filter_none(
            session=session,
            filter=filter,
            remove=remove,
        )

        return await self.request(Route("GET", "/user"), payload)

    async def get_user_favorites(
        self,
        session: str,
        filter: list[str] | None = None,
        remove: list[str] | None = None,
        include: list[str] | None = None,
        description_type: str | None = None,
        playlist_type: str | None = None,
    ) -> list[dict]:
        payload: dict = dict_filter_none(
            session=session,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        return await self.request(Route("GET", "/user/favourites"), payload)

    async def add_user_favorite(self, session: str, title_id: int) -> dict:
        payload: dict = dict_filter_none(session=session, title_id=title_id)
        return await self.request(Route("PUT", "/user/favourites/add"), payload)

    async def remove_user_favorite(self, session: str, title_id: int) -> dict:
        payload: dict = dict_filter_none(session=session, title_id=title_id)
        return await self.request(Route("DELETE", "/user/favourites/remove"), payload)



__all__ = ("HTTPException", "NoArgumentsError")


class HTTPException(Exception):
    def __init__(self, code: int, message: str) -> None:
        super().__init__(f"HTTP error with code: {code}!\nMessage: {message}")


class NoArgumentsError(Exception):
    def __init__(self, *missed_args):
        super().__init__(f"Should be at least one argument of {', '.join(missed_args)}")



from typing import ClassVar, TypeVar

__all__ = ("T", "MISSING", "Absent")

T = TypeVar("T")


class _Missing:
    _instance: ClassVar["_Missing"] = None

    def __new__(cls):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls)
        return cls._instance

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __repr__(self):
        return "<anilibria.MISSING>"

    def __hash__(self):
        return 0

    def __bool__(self):
        return False


MISSING = _Missing()
Absent = T | _Missing



from typing import List

from .attrs_utils import define
from .enums import SeasonCode, StatusCode, TitleCodeType

__all__ = (
    "TitleNames",
    "TitleStatus",
    "Poster",
    "Posters",
    "TitleType",
    "TitleTeam",
    "Season",
    "Blocked",
    "Episodes",
    "HLS",
    "SerieSkips",
    "Episode",
    "RutubeEpisode",
    "Player",
    "Quality",
    "TorrentFile",
    "TorrentMetaData",
    "Torrent",
    "Torrents",
    "Title",
)


@define()
class TitleNames:
    """
    Объект, содержащий названия тайтла на различных языках.
    """

    ru: str | None = None
    "Название тайтла на русском языке"
    en: str | None = None
    "Название тайтла на английском языке"
    alternative: str | None = None
    "Название тайтла на альтернативном языке"


@define()
class TitleStatus:
    """
    Объект статуса тайтла
    """

    string: str | None = None
    "Представление статуса в виде строки"
    code: StatusCode | None = None
    "Код статуса"


@define()
class Poster:
    """
    Объект с моделью постера
    """

    url: str | None = None
    "Относительная ссылка на постер"
    raw_base64_file: str | None = None
    "Закодированный в base64 постер"

    @property
    def full_url(self) -> str | None:
        """
        Возвращает полную ссылку на постер
        """
        if self.url is not None:
            return f"https://anilibria.tv{self.url}"


@define()
class Posters:
    """
    Объект с разными размерами постера
    """

    small: Poster | None = None
    "Постер маленького размера"
    medium: Poster | None = None
    "Постер среднего размера"
    original: Poster | None = None
    "Постер оригинального размера"


@define()
class TitleType:
    """
    Объект с информацией о типе тайтла.
    """

    full_string: str | None = None
    "Полная информация о типе в виде строки"
    code: TitleCodeType | None = None
    "Код типа"
    string: str | None = None
    "Тип тайтла в виде строки"
    episodes: int | None = None
    "Количество серий"
    length: str | None = None
    "Длина серии"


@define()
class TitleTeam:
    """
    Объект с участниками, которые принимали участие в переводе тайтла.
    """

    voice: List[str] | None = None
    "Участники, работавшие над озвучкой"
    translator: List[str] | None = None
    "Участники, работавшие над переводом"
    editing: List[str] | None = None
    "Участники, работавшие над субтитрами"
    decor: List[str] | None = None
    "Участники, работавшие над оформлением"
    timing: List[str] | None = None
    "Участники, работавшие над таймингом"


@define()
class Season:
    """
    Объект с информацией о сезоне тайтла.
    """

    string: str | None = None
    "Название сезона"
    code: SeasonCode | None = None
    "Код сезона"
    year: int | None = None
    "Год выпуска"
    week_day: int | None = None
    "День недели"


@define()
class Blocked:
    """
    Объект с информацией о статусе блокировки тайтла.
    """

    blocked: bool | None = False
    "Тайтл заблокирован на территории РФ"
    bakanim: bool | None = False
    "Тайлтл заблокирован из-за жалобы Wakanim"


@define()
class Episodes:
    """
    Объект с информацией о количестве серий
    """

    first: int | None = None
    "Номер первой серии"
    last: int | None = None
    "Номер последней серии"
    string: str | None = None
    "Представление количества серий в виде строки"


@define()
class HLS:
    """
    Объект, содержащий ссылки на серии в различных качествах.

    ..warning
        Ссылки являются относительными и не содержат домена!
    """

    fhd: str | None = None
    "Ссылка на видео в Full HD качестве"
    hd: str | None = None
    "Ссылка на видео в HD качестве"
    sd: str | None = None
    "Ссылка на видео в SD качестве"


@define()
class SerieSkips:
    """
    Объект с таймкодами для пропуска опенинга и эндинга.
    """

    opening: List[str] | None = None
    "Таймкоды для опенинга"
    ending: List[str] | None = None
    "Таймкоды для эндинга"


@define()
class Episode:
    """
    Объект, содержащий информацию о серии.
    """

    episode: int | None = None
    "Номер серии"
    created_timestamp: int | None = None
    "Время создания/изменения в формате UNIX timestamp"
    hls: HLS | None = None
    "Ссылки на серию"
    preview: str | None = None
    "Ссылка на превью серии"
    skips: SerieSkips | None = None
    "Таймкоды на пропуски"

    # TODO: Добавить свойство 'created_at' с datetime


@define()
class RutubeEpisode:
    """
    Объект с информацией о серии в rutube
    """

    created_timestamp: int | None = None
    "Время создания/изменения в формате UNIX timestamp"
    rutube_id: str | None = None  # TODO: Возможно ли из айди собрать ссылку?
    "ID серии"
    episode: int | None = None
    "Номер серии"


@define()
class Player:
    """
    Объект с информацией о плеере и сериях.
    """

    alternative_player: str | None = None
    "Ссылка на альтернативный плеер"
    host: str | None = None
    "Имена предпочитаемых серверов для построения ссылок на поток и скачивание"
    episodes: Episodes | None = None
    "Количество вышедших серий"
    list: dict[str, Episode] | List[Episode] = None
    "Список релизов"
    rutube: dict[str, RutubeEpisode] | List[RutubeEpisode] = None
    "Список релизов на rutube"


@define()
class Quality:
    """
    Объект, содержащий информацию о разрешении, кодировщике и типе релиза
    """

    string: str | None = None
    "Полная информация о качестве"
    type: str | None = None  # TODO: Enum!
    "Тип релиза"
    resolution: str | None = None
    "Разрешение серии"
    encoder: str | None = None
    "Название кодировщика"
    lq_audio: bool | None = None
    "Используется ли аудио дорожка с пониженным битрейтом"


@define()
class TorrentFile:
    """
    Объект с информацией о торрент файле
    """

    file: str | None = None
    "Имя файла"
    size: int | None = None
    "Размер файла в байтах"
    offset: int | None = None
    "Смещение в байтах относительно предыдущего файла"


@define()
class TorrentMetaData:
    """
    Объект с метадатой о торренте
    """

    hash: str | None = None
    "Хеш торрент файла"
    name: str | None = None
    "Имя тайтла в торрент файле"
    announce: List[str] | None = None
    "Список трекеров"
    created_timestamp: int | None = None
    "Время создания торрента в UNIX timestamp"
    files_list: List[TorrentFile] | None = None
    "Список файлов в торренте"


@define()
class Torrent:
    """
    Объект с информацией о торренте
    """

    torrent_id: int | None = None
    "ID торрент файла"
    episodes: Episodes | None = None
    "Серии, содержащиеся в файле"
    quality: Quality | None = None
    "Информация о разрешении, кодировщике и типе релиза"
    leechers: int | None = None
    "Количество личей"
    seeders: int | None = None
    "Количество сидов"
    downloads: int | None = None
    "Количество загрузок"
    total_size: int | None = None
    "Размер файлов в торренте в байтах"
    url: str | None = None
    "Ссылка на торрент без домена"
    uploaded_timestamp: int | None = None
    "Время загрузки домена в формате UNIX timestamp"
    metadata: TorrentMetaData | None = None
    "Метаданные торрент файла"
    raw_base64_file: str | None = None
    "Торрент файл в формате base64"
    hash: str | None = None
    "Хэш торрент файла"
    magnet: str | None = None
    "Магнитная ссылка для скачивания торрента"


@define()
class Torrents:
    """
    Модель со списком торрентов и информации о сериях.
    """

    episodes: Episodes | None = None
    "Серии, содержащиеся в файле"
    list: List[Torrent] | None = None
    "Список с информацией о торрент файлах"


@define()
class TitleDescription:
    """
    Объект с описанием тайтла
    """

    html: str | None = None
    "Описание тайтла в виде html"
    no_view_order: str | None = None
    "Описание тайтла в виде текста без дополнительного форматирования и порядка просмотра"
    plain: str | None = None
    "Описание тайтла без дополнительного форматирования"


@define()
class Title:
    """
    Объект тайтла
    """

    id: int | None = None
    "ID тайтла"
    code: str | None = None
    "Код тайтла"
    names: TitleNames | None = None
    "Названия тайтла"
    announce: str | None = None
    "Объявление для тайтла"
    status: TitleStatus | None = None
    "Статус тайтла"
    posters: Posters | None = None
    "Информация о постерах"
    updated: int | None = None
    "Время последнего обновления тайтла в формате UNIX timestamp"
    last_change: int | None = None
    "Время последнего изменения тайтла в формате UNIX timestamp"
    type: TitleType | None = None
    "Информация о типе тайтла"
    genres: List[str] | None = None
    "Список жанров тайтла"
    team: TitleTeam | None = None
    "Члены команды, работавшие над тайтлом"
    season: Season | None = None
    "Информация о сезоне"
    description: str | None = None
    "Описание тайтла"
    in_favorites: int | None = None
    "Сколько раз тайтл добавили в избранное"
    blocked: Blocked | None = None
    "Информация о блокировке тайтла"
    player: Player | None = None
    "Информация о плеере"
    torrents: Torrents | None = None
    "Информация о торрентах"

    @property
    def url(self) -> str:
        """
        Возвращает полную ссылку на тайтл
        """
        return f"https://anilibria.tv/release/{self.code}.html"



from .dispatch import *  # noqa: F401 F403
from .error import *  # noqa: F401 F403
from .gateway import *  # noqa: F401 F403
from .http import *  # noqa: F401 F403
from .models import *  # noqa: F401 F403



from enum import Enum

from .typings import MISSING

__all__ = ("dict_filter_none", "dict_filter_missing", "prepare_payload")


def dict_filter_none(**kwargs) -> dict:
    return {key: value for key, value in kwargs.items() if value is not None}


def dict_filter_missing(data: dict = None, /, **kwargs):
    _data: dict = data or kwargs

    return {key: value for key, value in _data.items() if value is not MISSING}


def prepare_payload(data: dict):
    for key, value in data.items():
        if isinstance(value, list):
            data[key] = ",".join(str(_) for _ in value)
        elif isinstance(value, Enum):
            data[key] = value.value



from typing import Generic, List, TypeVar

from .attrs_utils import define
from .title import Title

T = TypeVar("T")

__all__ = ("Schedule", "YouTubeVideo", "SeedStats", "Pagination", "ListPagination", "User")


@define()
class Schedule:
    day: int
    "День недели"
    list: list[Title]
    "Список тайтлов"


@define()
class YouTubeVideo:
    id: int
    """Айди записи в базе"""
    title: str
    "Название youtube видео"
    image: str
    "Ссылка на превью к видео"
    youtube_id: str
    "Айди видео в youtube"
    timestamp: int
    "Время создания в формате UNIX timestamp"
    comments: int
    "Количество комментариев у видео"
    views: int
    "Количество просмотров у видео"

    @property
    def url(self) -> str:
        """Возвращает полную youtube ссылку на видео"""
        return f"https://youtu.be/{self.youtube_id}"


@define()
class SeedStats:
    user: str
    "Имя пользователя"
    downloaded: int
    "Сколько байт было скачано"
    uploaded: int
    "Сколько байт было загружено"


@define()
class Pagination:
    current_page: int
    "Текущая страница"
    pages: int
    "Всего страниц"
    items_per_page: int
    "Количество элементов на странице"
    total_items: int
    "Всего элементов"


@define()
class ListPagination(Generic[T]):
    pagination: Pagination
    "Объект пагинации"
    list: List[T]
    "Список объектов определённого типа"


@define()
class User:
    login: str | None = None
    "Логин пользователя"
    nickname: str | None = None
    "Ник пользователя"
    email: str | None = None
    "Электронная почта пользователя"
    avatar_original: str | None = None
    "Путь к аватару пользователя"
    avatar_thumbnail: str | None = None
    "Путь к превью аватара пользователя"
    vk_id: str | None = None
    "Айди аккаунта Вконтакте"
    patreon_id: str | None = None
    "Айди аккаунта в Patreon"



from ...models import Episode, Title
from ...models.attrs_utils import define
from .base import BaseEvent

__all__ = (
    "Connect",
    "TitleEpisode",
)


@define()
class Connect(BaseEvent):
    """
    Модель для события ``on_connect``. Вызывается при подключении к АПИ анилибрии. Может вызываться несколько раз.

    .. code-block:: python

      @client.on(Connect)
      async def connected(event: Connect):
          ...
    """

    api_version: str
    "Версия АПИ анилибрии"


@define()
class TitleEpisode(BaseEvent):
    """
    Модель для ивента `on_title_episode`. Вызывается при полной загрузки эпизода на сервер.

    .. code-block:: python

      @client.on(TitleEpisode)
      async def on_title_episode(event: TitleEpisode):
          ...
    """

    title: Title
    "Объект тайтла"
    episode: Episode
    "Объект загруженного эпизода"



import asyncio
from logging import DEBUG, basicConfig, getLogger
from typing import Callable, Coroutine, Type

from aiohttp import WSServerHandshakeError

from ..api.error import NoArgumentsError
from ..api.gateway.client import GatewayClient
from ..api.gateway.events import BaseEvent, EventType, PlaylistUpdate, TitleEpisode
from ..api.http.client import HTTPClient
from ..api.models import (
    DescriptionType,
    Include,
    ListPagination,
    PlaylistType,
    RSSType,
    Schedule,
    SeedStats,
    Title,
    TitleTeam,
    User,
    YouTubeVideo,
)
from ..api.models.cattrs_utils import converter
from ..utils.serializer import dict_filter_missing
from ..utils.typings import MISSING, Absent

log = getLogger("anilibria.client")
__all__ = ("AniLibriaClient",)


class AniLibriaClient:
    """
    Основной клиент для взаимодействия с API anilibria.tv.
    """

    def __init__(self, *, proxy: str | None = None, logging: bool | int | None = None) -> None:
        self._http: HTTPClient = HTTPClient(proxy=proxy)
        self._websocket: GatewayClient = GatewayClient(http=self._http)

        if logging is not None:
            if logging is True:
                basicConfig(level=DEBUG)
            else:
                basicConfig(level=logging)

        self.event(self._on_playlist_update, name="on_playlist_update")

        self._loop = asyncio.get_event_loop()

    async def _on_playlist_update(self, event: PlaylistUpdate):
        # Убеждаемся, что ивент затрагивает обновление эпизода, а не другие данные
        if not event.updated_episode or not event.updated_episode.hls:
            return
        # Проверяем, что хотя бы одно из трёх качеств видео отсутствует, иначе это перезалив
        hls = event.updated_episode.hls
        if not hls.sd or not hls.hd or not hls.fhd:
            return
        # Смотрим предыдущие значения:
        # - списка эпизодов
        if (playlist := event.diff.get("list")) is None:
            return
        # - Конкретный эпизод
        if (episode := playlist.get(str(event.updated_episode.episode))) is None:
            return
        # - значения hls эпизода
        if (previous_hls := episode.get("hls")) is None:
            return
        # Если есть предыдущее значение, значит это перезалив
        if all(v is not None for v in previous_hls.values()):
            return

        title = await self.get_title(id=event.id)

        self._websocket.dispatch.call(
            "on_title_episode", TitleEpisode(title=title, episode=event.updated_episode)
        )

    def on_startup(self, coro: Callable[..., Coroutine] = MISSING):
        """
        Декоратор для прослушивания события ``on_startup``. Вызывается только один раз при запуске клиента.

        .. code-block:: python

           @client.on_startup()  # Можно без скобок
           async def start():  # Единственное событие, при котором функция ничего не должна принимать.
               ...

        :param Callable[..., Coroutine] coro: Функция, которая будет вызываться.
        """

        def wrapper(coro: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
            self._websocket.dispatch.register("on_startup", coro)

            return coro

        if coro is not MISSING:
            return wrapper(coro)

        return wrapper

    def on(self, event: Type[BaseEvent]):
        """
        Декоратор для прослушивания событий. Принимает класс события.

        .. code-block:: python

           @client.on(PlaylistUpdate)
           async def name_you_want(event: PlaylistUpdate):
               ...

        :param event: Класс ивента
        """

        def wrapper(coro: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
            event_name: str = "on_" + EventType(event).name.lower()
            self._websocket.dispatch.register(event_name, coro)

            return coro

        return wrapper

    def listen(self, coro: Callable[..., Coroutine] = MISSING, *, name: str = MISSING):
        """
        Декоратор для прослушивания событий. Принимает названия события.

        Примеры использования:

        .. code-block:: python

           @client.listen
           async def on_playlist_update(event: PlaylistUpdate):
               ...

           @client.listen()
           async def on_playlist_update(event: PlaylistUpdate):
               ...

           @client.listen(name="on_playlist_update")
           async def name_you_want(event: PlaylistUpdate):
               ...

        :param Callable[..., Coroutine] coro: Функция, которая будет вызываться.
        :param str name: Название ивента. Например: on_title_update.
        """

        def decorator(coro: Callable[..., Coroutine]):
            self._websocket.dispatch.register(name or coro.__name__, coro)
            return coro

        if coro is not MISSING:
            return decorator(coro)

        return decorator

    def event(self, coro: Callable[..., Coroutine] = MISSING, *, name: str = MISSING):
        """
        Декоратор для прослушивания событий. Принимает названия события.
        Алиас для :meth:`.listen`:

        Примеры использования:

        .. code-block:: python

           @client.listen
           async def on_playlist_update(event: PlaylistUpdate):
               ...

           @client.listen()
           async def on_playlist_update(event: PlaylistUpdate):
               ...

           @client.listen(name="on_playlist_update")
           async def name_you_want(event: PlaylistUpdate):
               ...

        :param Callable[..., Coroutine] coro: Функция, которая будет вызываться.
        :param str name: Название ивента. Например: on_title_update.
        """

        return self.listen(coro, name=name)

    async def subscribe(self, subscribe: dict, filter: str = MISSING, remove: str = MISSING):
        """
        По умолчанию клиент получает все возможные уведомления от API.
        Но можно подписаться на определённые ивенты, или ивенты с каким-то фильтром

        .. code-block:: python

           await client.subscribe(
               {
                   "title_update": {
                       "title": {
                           "season": {
                               "year": 2022
                           }
                       }
                   }
               }
           )

        :param dict subscribe: Данные о подписке. Здесь может быть всё то, что принимает веб сокет.
        :param str filter: То, что должно быть включено в подписку.
        :param str remove: То, что нужно удалить из подписки.
        """
        data = {"subscribe": subscribe}
        if filter is not MISSING:
            data["filter"] = filter
        if remove is not MISSING:
            data["remove"] = remove

        await self._websocket.subscribe(data)

    async def login(self, mail: str, password: str) -> str:
        """
        Входит в аккаунт.

        .. warning::
           Если запрос идёт из РФ, то для использования необходим VPN или proxy!

        :param str mail: Логин или эл.почта
        :param str password: Пароль
        :return: ID сессии
        """
        data = await self._http.login(mail, password)
        return data.get("sessionId")

    async def get_title(
        self,
        id: Absent[int] = MISSING,
        code: Absent[str] = MISSING,
        torrent_id: Absent[int] = MISSING,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
    ) -> Title:
        """
        Возвращает тайтл с заданными параметрами.

        :param Absent[int] id: ID тайтла.
        :param Absent[str] code: Код тайтла.
        :param Absent[int] torrent_id: ID торрента
        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        """
        if id is MISSING and code is MISSING:
            raise NoArgumentsError("id", "code")

        payload: dict = dict_filter_missing(
            id=id,
            code=code,
            torrent_id=torrent_id,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
        )

        data = await self._http.get_title(**payload)
        return converter.structure(data, Title)

    async def get_titles(
        self,
        id_list: Absent[list[int]] = MISSING,
        code_list: Absent[list[str]] = MISSING,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список тайтлов с заданными параметрами.

        :param Absent[list[int]] id_list: Список с ID тайтлами
        :param Absent[list[str]] code_list: Список с кодами тайтлов.
        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        if id_list is MISSING and code_list is MISSING:
            raise NoArgumentsError("id_list", "code_list")

        payload: dict = dict_filter_missing(
            id_list=id_list,
            code_list=code_list,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            page=page,
            items_per_page=items_per_page,
        )

        data = await self._http.get_titles(**payload)

        return converter.structure(data, ListPagination)

    async def get_updates(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        since: Absent[int] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список последних обновлений тайтлов с заданными параметрами.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[int] since: Список тайтлов, у которых время обновления больше указанного timestamp
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        :param Absent[int] after: Удаляет первые n записей из выдачи
        :param Absent[int] limit: Количество объектов в ответе. По умолчанию 5
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_updates(**payload)
        return converter.structure(data, ListPagination)

    async def get_changes(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        since: Absent[int] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список последних обновлений тайтлов с заданными параметрами.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[int] since: Список тайтлов, у которых время обновления больше указанного timestamp
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[int] after: Удаляет первые n записей из выдачи
        :param Absent[int] limit: Количество объектов в ответе. По умолчанию 5
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_changes(**payload)
        return converter.structure(data, ListPagination)

    async def get_schedule(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        days: Absent[list[int]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
    ) -> list[Schedule]:
        """
        Возвращает список последних обновлений тайтлов с заданными параметрами по дням.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[list[int]] days: Список дней недели, на которые нужно расписание
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            days=days,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        data = await self._http.get_schedule(**payload)
        return converter.structure(data, list[Schedule])

    async def get_random_title(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
    ) -> Title:
        """
        Возвращает рандомный тайтл с заданными параметрами.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
        )
        data = await self._http.get_random_title(**payload)
        return converter.structure(data, Title)

    async def get_youtube(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        since: Absent[int] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[YouTubeVideo]:
        """
        Возвращает список youtube видео в хронологическом порядке с заданными параметрами.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[int] since: Список тайтлов, у которых время обновления больше указанного timestamp
        :param Absent[int] after: Удаляет первые n записей из выдачи
        :param Absent[int] limit: Количество объектов в ответе. По умолчанию 5
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_youtube(**payload)
        return converter.structure(data, ListPagination)

    async def get_feed(
        self,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        since: Absent[int] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title | YouTubeVideo]:
        """
        Возвращает список тайтлов и youtube видео в хронологическом порядке с заданными параметрами.

        :param Absent[list[str]] filter: То, что должно быть в ответе.
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[int] since: Список тайтлов, у которых время обновления больше указанного timestamp
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        :param Absent[int] after: Удаляет первые n записей из выдачи
        :param Absent[int] limit: Количество объектов в ответе. По умолчанию 5
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            filter=filter,
            remove=remove,
            include=include,
            since=since,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_feed(**payload)

        data["list"]: list
        video: dict

        for index, video in enumerate(data["list"]):
            if title := video.get("title"):
                data["list"][index] = converter.structure(title, Title)
            else:
                data["list"][index] = converter.structure(data["youtube"], YouTubeVideo)

        return converter.structure(data, ListPagination)

    async def get_years(self) -> list[int]:
        """
        Возвращает список годов выхода доступных тайтлов отсортированный по возрастанию.
        """
        return await self._http.get_years()

    async def get_genres(self, sorting_type: int = 0) -> list[str]:
        """
        Возвращает список жанров доступных тайтлов отсортированный по алфавиту.

        :param int sorting_type: Тип сортировки элементов.
        """
        return await self._http.get_genres(sorting_type=sorting_type)

    async def get_team(self) -> TitleTeam:
        """
        Возвращает участников команды когда-либо существовавших на проекте.
        """
        data = await self._http.get_team()
        return converter.structure(data, TitleTeam)

    async def get_seed_stats(
        self,
        users: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        after: Absent[int] = MISSING,
        sort_by: Absent[str] = MISSING,
        order: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[SeedStats]:
        """
        Возвращает топ пользователей по количеству загруженного и скачанно через торрент трекер anilibria.

        :param Absent[list[str]] users: Статистика по имени пользователя
        :param Absent[list[str]] remove: То, чего не должно быть в ответе.
        :param Absent[list[Include]] include: Список типов файлов которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list)
        :param Absent[int] after: Удаляет первые n записей из выдачи.
        :param Absent[str] sort_by: По какому полю производить сортировку, допустимые значения: downloaded, uploaded, user
        :param Absent[int] order: Направление сортировки 0 - DESC, 1 - ASC.
        :param Absent[int] limit: Количество объектов в ответе. По умолчанию 5
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            users=users,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            sort_by=sort_by,
            order=order,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_seed_stats(**payload)
        return converter.structure(data, ListPagination)

    async def get_rss(
        self,
        rss_type: Absent[RSSType] = MISSING,
        session_id: Absent[str] = MISSING,
        since: Absent[int] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
    ) -> str:
        """
        Возвращает список обновлений на сайте в одном из форматов RSS ленты.

        :param Absent[RSSType] rss_type: Предпочитаемый формат вывода
        :param Absent[str] session_id: Уникальный идентификатор сессии пользователя
        :param Absent[int] since: Список тайтлов у которых время обновления больше указанного timestamp
        :param Absent[int] after: Удаляет первые n записей из выдачи
        :param Absent[int] limit: Количество объектов в ответе
        """
        payload: dict = dict_filter_missing(
            rss_type=rss_type, session_id=session_id, since=since, after=after, limit=limit
        )

        return await self._http.get_rss(**payload)

    async def search_titles(
        self,
        search: Absent[list[str]] = MISSING,
        year: Absent[list[str | int]] = MISSING,
        season_code: Absent[list[str]] = MISSING,
        genres: Absent[list[str]] = MISSING,
        team: Absent[list[str]] = MISSING,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        after: Absent[int] = MISSING,
        limit: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список тайтлов, найденных по фильтрам.

        :param Absent[list[str]] search: Поиск по именам и описанию.
        :param Absent[list[str | int]] year: Список годов выхода.
        :param Absent[list[str]] season_code: Список сезонов.
        :param Absent[list[str]] genres: Список жанров.
        :param Absent[list[str]] team: Ники участников, работавшие над тайтлом.
        :param Absent[list[str]] filter: Список значений, которые будут в ответе.
        :param Absent[list[str]] remove: Список значений, которые будут удалены из ответа.
        :param Absent[list[Include]] include: Список типов файлов, которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list).
        :param Absent[int] after: Удаляет первые n записей из выдачи.
        :param Absent[int] limit: Количество объектов в ответе.
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            search=search,
            year=year,
            season_code=season_code,
            genres=genres,
            team=team,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            limit=limit,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.search_titles(**payload)
        return converter.structure(data, list[Title])

    async def advanced_search(
        self,
        query: str,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        after: Absent[int] = MISSING,
        order_by: Absent[str] = MISSING,
        limit: Absent[int] = MISSING,
        sort_direction: Absent[int] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список тайтлов, найденных по фильтрам.

        :param str query: Запрос для поиска. Может быть название тайтла.
        :param Absent[list[str]] filter: Список значений, которые будут в ответе.
        :param Absent[list[str]] remove: Список значений, которые будут удалены из ответа.
        :param Absent[list[Include]] include: Список типов файлов, которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий.
        :param Absent[int] after: Удаляет первые n записей из выдачи.
        :param Absent[str] order_by: Ключ по которому будет происходить сортировка результатов
        :param Absent[int] limit: Количество объектов в ответе.
        :param sort_direction: Направление сортировки. 0 - По возрастанию, 1 - По убыванию
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            query=query,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            after=after,
            order_by=order_by,
            limit=limit,
            sort_direction=sort_direction,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.advanced_search(**payload)
        return converter.structure(data, ListPagination)

    async def get_user(
        self,
        session_id: int,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
    ) -> User:
        """
        Возвращает объект пользователя по текущей сессии.

        :param session_id: Айди сессии.
        :param Absent[list[str]] filter: Список значений, которые будут в ответе.
        :param Absent[list[str]] remove: Список значений, которые будут удалены из ответа.
        """
        payload = dict_filter_missing(session_id=session_id, filter=filter, remove=remove)
        data = await self._http.get_user(**payload)
        return converter.structure(data, User)

    async def get_user_favorites_titles(
        self,
        session_id: str,
        filter: Absent[list[str]] = MISSING,
        remove: Absent[list[str]] = MISSING,
        include: Absent[list[Include]] = MISSING,
        description_type: Absent[DescriptionType] = MISSING,
        playlist_type: Absent[PlaylistType] = MISSING,
        page: Absent[int] = MISSING,
        items_per_page: Absent[int] = MISSING,
    ) -> ListPagination[Title]:
        """
        Возвращает список избранных тайтлов пользователя

        :param str session_id: ID сессии.
        :param Absent[list[str]] filter: Список значений, которые будут в ответе.
        :param Absent[list[str]] remove: Список значений, которые будут удалены из ответа.
        :param Absent[list[Include]] include: Список типов файлов, которые будут возвращены в виде base64 строки
        :param Absent[DescriptionType] description_type: Тип получаемого описания.
        :param Absent[PlaylistType] playlist_type: Формат получаемого списка серий. Словарь(object) или список(list).
        :param Absent[int] page: Номер страницы. По умолчанию 1
        :param Absent[int] items_per_page: Количество элементов на одной странице.
        """
        payload = dict_filter_missing(
            session=session_id,
            filter=filter,
            remove=remove,
            include=include,
            description_type=description_type,
            playlist_type=playlist_type,
            page=page,
            items_per_page=items_per_page,
        )
        data = await self._http.get_user_favorites(**payload)
        return converter.structure(data, ListPagination)

    async def add_user_favorite_title(self, session_id: str, title_id: int):
        """
        Добавляет тайтл в список избранных

        :param str session_id: ID сессии.
        :param int title_id: ID тайтла.
        """
        await self._http.add_user_favorite(session=session_id, title_id=title_id)

    async def remove_user_favorite_title(self, session_id: str, title_id: int):
        """
        Удаляет тайтл из списка избранных

        :param str session_id: ID сессии.
        :param int title_id: ID тайтла.
        """
        await self._http.remove_user_favorite(session=session_id, title_id=title_id)

    async def astart(self, *, auto_reconnect: bool = True):
        """
        Запускает клиент асинхронно
        """

        while True:
            try:
                await self._websocket.start()
            except WSServerHandshakeError as error:
                if auto_reconnect:
                    log.debug("Websocket disconnected. Reconnecting...")
                    continue
                raise error from error
            except (KeyboardInterrupt, asyncio.CancelledError):
                await self.close()
                break

    def start(self, *, auto_reconnect: bool = True):
        """
        Запускает клиент.
        """

        try:
            self._loop.run_until_complete(self.astart(auto_reconnect=auto_reconnect))
        except KeyboardInterrupt:
            self._loop.run_until_complete(self.close())

    def startwith(self, coro: Coroutine, *, auto_reconnect: bool = True):
        """
        Запускает основной клиент вместе с корутиной.

        :param coro: Корутина для запуска
        :param auto_reconnect: Нужно ли перезапускать клиент после ошибки сервера анилибрии?
        """
        gather = asyncio.gather(
            self._loop.create_task(self.astart(auto_reconnect=auto_reconnect)),
            self._loop.create_task(coro),
        )
        self._loop.run_until_complete(gather)

    async def close(self):
        """
        Закрывает клиент.
        """
        if self._http.session and not self._http.session.closed:
            await self._http.session.close()
            await self._websocket.close()



from enum import Enum
from typing import Any, Type

__all__ = (
    "StrEnum",
    "IntEnum",
    "StatusCode",
    "TitleCodeType",
    "SeasonCode",
    "RSSType",
    "DescriptionType",
    "Include",
    "PlaylistType",
)


def _cursed_enum(cls: Type[Enum], obj: type, value: Any) -> Enum:
    new = obj.__new__(cls)  # type: ignore
    new._name_ = f"UNKNOWN: {value}"
    new._value_ = value

    return cls._value2member_map_.setdefault(value, new)


class IntEnum(int, Enum):
    """Enum where members must be ints"""

    @classmethod
    def _missing_(cls, value: int) -> Enum:
        return _cursed_enum(cls, int, value)


class StrEnum(str, Enum):
    """Enum where members must be strings"""

    @classmethod
    def _missing_(cls, value: str) -> Enum:
        return _cursed_enum(cls, str, value)

    def __str__(self):
        return self.value


class StatusCode(IntEnum):
    AT_WORK = 1
    "В работе"
    FINISHED = 2
    "Закончен"
    HIDDEN = 3
    "Скрыт"
    NO_ONGOING = 4
    "Неонгоинг"


class TitleCodeType(IntEnum):
    FILM = 0
    TV = 1
    OVA = 2
    ONA = 3
    SPECIAL = 4
    WEB = 5


class SeasonCode(IntEnum):
    """
    Представляет код сезона
    """

    UNKNOWN = 0
    "Неизвестный сезон"
    WINTER = 1
    "Зимний сезон"
    SPRING = 2
    "Весенний сезон"
    SUMMER = 3
    "Летний сезон"
    AUTUMN = FALL = 4
    "Осенний сезон"


class RSSType(StrEnum):
    """
    Представляет тип ответа на ``client.get_rss()``
    """

    RSS = "rss"
    ATOM = "atom"
    JSON = "json"


class DescriptionType(StrEnum):
    """
    Представляет тип описания тайтла
    """

    HTML = "html"
    PLAIN = "plain"
    NO_VIEW_ORDER = "no_view_order"


class Include(StrEnum):
    """
    Представляет запрашиваемый контент, который недоступен по умолчанию
    """

    RAW_POSTER = "raw_poster"
    RAW_TORRENT = "raw_torrent"
    TORRENT_META = "torrent_meta"


class PlaylistType(StrEnum):
    """
    Представляет тип плейлиста
    """

    ARRAY = LIST = "array"
    "Плейлист в виде списка"
    OBJECT = DICT = "object"
    "Плейлист в виде словаря"


class QualityType(StrEnum):
    WEBRip = "WEBRip"



from .attrs_utils import *  # noqa: F401 F403
from .cattrs_utils import *  # noqa: F401 F403
from .enums import *  # noqa: F401 F403
from .misc import *  # noqa: F401 F403
from .title import *  # noqa: F401 F403



import asyncio
from logging import getLogger

from aiohttp import ClientWebSocketResponse, WSMessage, WSMsgType
from orjson import dumps, loads

from ...const import __api_url__
from ..dispatch import Dispatch
from ..http import HTTPClient
from ..models.cattrs_utils import converter
from .events import Connect, EventType, Subscription

log = getLogger("anilibria.gateway")
URL = f"wss://{__api_url__}/ws/"
__all__ = ("GatewayClient",)


class GatewayClient:
    def __init__(self, http: HTTPClient):
        self._connection: ClientWebSocketResponse | None = None
        self._closed: bool = False
        self._stopped: bool = False

        self._http: HTTPClient = http
        self.dispatch: Dispatch = Dispatch()

        self._started_up: bool = False

    @property
    def loop(self):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()

        return loop

    async def start(self):
        await self.connect()

    async def close(self):
        if self._connection is not None:
            await self._connection.close()

    async def connect(self):

        if (session := self._http.session) is None or session.closed:
            session = await self._http.create_session()

        async with session.ws_connect(URL) as self._connection:
            self._closed = self._connection.closed

            if not self._started_up:
                self.dispatch.call("on_startup")
                self._started_up = True

            data = await self._receive_data()

            # Information about opened connection
            if data.get("connection") == "success":
                self.dispatch.call("on_connect", converter.structure(data, Connect))

            while not self._closed:
                data = await self._receive_data()

                # Possible only when connection was closed
                if isinstance(data, WSMessage):
                    return

                self._track_data(data)

    async def _receive_data(self) -> dict | WSMessage:
        response = await self._connection.receive()

        if response.type is WSMsgType.CLOSING:
            self._closed = True
            return response

        return response.json(loads=loads)

    def _track_data(self, data: dict):
        type: str
        if not (type := data.get("type")):
            return self._track_unknown_event(data)

        event_model = EventType[type.upper()].value

        if event_model is None:
            return log.warning(f"Received a not excepted event `{type}`!")

        log.debug(f"Received {type} event {data}")

        obj = converter.structure(data["data"], event_model)
        self.dispatch.call(f"on_{type}", obj)

    def _track_unknown_event(self, data: dict):
        if "subscribe" in data:
            self.dispatch.call("on_subscription", converter.structure(data, Subscription))

    async def _send_message(self, data: dict):
        await self._connection.send_bytes(dumps(data))

    async def subscribe(self, data: dict):
        await self._send_message(data)



from codecs import open
from pathlib import Path

import tomli
from setuptools import find_packages, setup

with open("pyproject.toml", "rb") as f:
    pyproject = tomli.load(f)


def get_long_description() -> str:
    return (Path(__file__).parent / "README.rst").read_text()


def get_requirements(filename: str) -> list[str]:
    return (Path(__file__).parent / filename).read_text().splitlines()


extra_requirements = {
    "readthedocs": [
        "furo",
        "readthedocs-sphinx-search",
        "Sphinx",
        "sphinx-hoverxref",
    ]
}


setup(
    name=pyproject["tool"]["poetry"]["name"],
    version=pyproject["tool"]["poetry"]["version"],
    author="Damego",
    author_email="danyabatueff@gmail.com",
    description=pyproject["tool"]["poetry"]["description"],
    extras_require=extra_requirements,
    include_package_data=True,
    install_requires=[
        "aiohttp==3.8.3",
        "attrs==22.2.0",
        "cattrs==22.2.0",
        "orjson==3.8.5",
        "tomli==2.0.1",
    ],
    license="GPL-3.0 License",
    long_description_content_type="text/markdown",
    long_description=get_long_description(),
    url="https://github.com/Damego/anilibria.py",
    packages=find_packages(),
    python_requires=">=3.10",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)



import asyncio
from collections import defaultdict
from logging import getLogger
from typing import Callable, Coroutine, Dict, List

log = getLogger("anilibria.dispatch")
__all__ = ("Dispatch",)


class Dispatch:
    def __init__(self) -> None:
        self._registered_events: Dict[str, List[Callable[..., Coroutine]]] = defaultdict(list)

    @staticmethod
    async def _call(coro: Callable[..., Coroutine], *args):
        try:
            await coro(*args)
        except Exception:  # noqa
            log.exception("")

    def call(self, name: str, *args):
        log.debug(f"Dispatching event {name}")

        for coro in self._registered_events.get(name, []):
            asyncio.create_task(self._call(coro, *args))

    def register(self, name: str, coro: Callable[..., Coroutine]):
        self._registered_events[name].append(coro)

        log.debug(
            f"Added coro to {name} event. Total coros for this event: {self._registered_events[name]}"
        )



from functools import wraps
from logging import getLogger

import attrs

__all__ = ("define",)
log = getLogger("anilibria.attrs")


define_defaults = dict(kw_only=True)


@wraps(attrs.define)
def define(**kwargs):
    return attrs.define(**kwargs, **define_defaults)



from logging import getLogger

from aiohttp import ClientResponse, ClientSession
from orjson import JSONDecodeError, loads

from ...utils.serializer import prepare_payload
from ..error import HTTPException
from .route import Route

log = getLogger("anilibria.request")
__all__ = ("Request",)


class Request:
    def __init__(self, proxy: str | None = None) -> None:
        self.proxy: str | None = proxy
        self.session: ClientSession = None  # noqa

    async def create_session(self) -> ClientSession:
        if self.session is None or self.session.closed:
            self.session = ClientSession()

        return self.session

    async def request(self, route: Route, params: dict = None, **kwargs):
        await self.create_session()

        if params is not None:
            prepare_payload(params)

        if self.proxy is not None:
            kwargs["proxy"] = self.proxy

        log.debug(
            f"Send {route.method} request to {route.endpoint} endpoint with params: {params} and kwargs: {kwargs}"
        )

        async with self.session.request(
            route.method, route.url, params=params, **kwargs
        ) as response:
            data = await self._get_data(response)

            log.debug(f"Got response from request {data}")

            self._catch_error(data)

            return data

    @staticmethod
    async def _get_data(response: ClientResponse) -> dict | str:
        try:
            data = await response.json(loads=loads)
        except JSONDecodeError:  # Could be RSS
            data = await response.text()

        return data

    @staticmethod
    def _catch_error(data: dict):
        if not isinstance(data, dict):
            return

        if error := data.get("error"):
            raise HTTPException(error["code"], error["message"])

        if data.get("err"):
            raise HTTPException(0, data["mes"])
